#!/usr/bin/env python
# coding: utf-8
"""Variants of Delhommeau's method for the computation of the Green function."""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging
from functools import lru_cache

import numpy as np

from capytaine.tools.prony_decomposition import exponential_decomposition, error_exponential_decomposition

from capytaine.green_functions.abstract_green_function import AbstractGreenFunction
import capytaine.green_functions.Delhommeau_f90 as Delhommeau_f90
import capytaine.green_functions.XieDelhommeau_f90 as XieDelhommeau_f90

LOG = logging.getLogger(__name__)


class Delhommeau(AbstractGreenFunction):
    """The Green function as implemented in Aquadyn and Nemoh.

    Parameters
    ----------
    tabulation_nr: int, optional
        Number of tabulation points for horizontal distance.
        Default: 328, as in Nemoh 2.
    tabulation_nz: int, optional
        Number of tabulation points for vertical distance.
        Default: 46, as in Nemoh 2.
    tabulation_nb_integration_points: int, optional
        Number of points for the numerical integration w.r.t. :math:`theta` of Delhommeau's integrals
        Default: 251, as in Nemoh 2.
    finite_depth_prony_decomposition_method: string, optional
        The implementation of the Prony decomposition used to compute the finite depth Green function.
        Accepted values: :code:`'fortran'` for Nemoh's implementation (by default), :code:`'python'` for an experimental Python implementation.
        See :func:`find_best_exponential_decomposition`.

    Attributes
    ----------
    tabulated_r_range: numpy.array of shape(tabulation_nr,)
    tabulated_z_range: numpy.array of shape(tabulation_nz,)
        Coordinates of the tabulation points.
    tabulated_integrals: numpy.array of shape (tabulation_nr, tabulation_nz, 2, 2)
        Tabulated Delhommeau integrals.
    """

    fortran_core = Delhommeau_f90

    def __init__(self, *,
                 tabulation_nr=328,
                 tabulation_nz=46,
                 tabulation_nb_integration_points=251,
                 finite_depth_prony_decomposition_method='fortran',
                 ):

        self.tabulated_r_range = self.fortran_core.delhommeau_integrals.default_r_spacing(tabulation_nr)
        self.tabulated_z_range = self.fortran_core.delhommeau_integrals.default_z_spacing(tabulation_nz)
        self.tabulated_integrals = self.fortran_core.delhommeau_integrals.construct_tabulation(
                self.tabulated_r_range, self.tabulated_z_range, tabulation_nb_integration_points
                )

        self.finite_depth_prony_decomposition_method = finite_depth_prony_decomposition_method

        self.exportable_settings = {
            'green_function': self.__class__.__name__,
            'tabulation_nr': tabulation_nr,
            'tabulation_nz': tabulation_nz,
            'tabulation_nb_integration_points': tabulation_nb_integration_points,
            'finite_depth_prony_decomposition_method': finite_depth_prony_decomposition_method,
        }

        self._hash = hash(self.exportable_settings.values())

    def __hash__(self):
        return self._hash

    @lru_cache(maxsize=128)
    def find_best_exponential_decomposition(self, dimensionless_omega, dimensionless_wavenumber):
        """Compute the decomposition of a part of the finite depth Green function as a sum of exponential functions.

        Two implementations are available: the legacy Fortran implementation from Nemoh and a newer one written in Python.
        For some still unexplained reasons, the two implementations do not always give the exact same result.
        Until the problem is better understood, the Fortran implementation is the default one, to ensure consistency with Nemoh.
        The Fortran version is also significantly faster...

        Results are cached.

        Parameters
        ----------
        dimensionless_omega: float
            dimensionless angular frequency: :math:`kh \\tanh (kh) = \\omega^2 h/g`
        dimensionless_wavenumber: float
            dimensionless wavenumber: :math:`kh`
        method: string, optional
            the implementation that should be used to compute the Prony decomposition

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            the amplitude and growth rates of the exponentials
        """

        LOG.debug(f"\tCompute Prony decomposition in finite depth Green function "
                  f"for dimless_omega=%.2e and dimless_wavenumber=%.2e",
                  dimensionless_omega, dimensionless_wavenumber)

        if self.finite_depth_prony_decomposition_method.lower() == 'python':
            # The function that will be approximated.
            @np.vectorize
            def f(x):
                return Delhommeau_f90.initialize_green_wave.ff(x, dimensionless_omega, dimensionless_wavenumber)

            # Try different increasing number of exponentials
            for n_exp in range(4, 31, 2):

                # The coefficients are computed on a resolution of 4*n_exp+1 ...
                X = np.linspace(-0.1, 20.0, 4*n_exp+1)
                a, lamda = exponential_decomposition(X, f(X), n_exp)

                # ... and they are evaluated on a finer discretization.
                X = np.linspace(-0.1, 20.0, 8*n_exp+1)
                if error_exponential_decomposition(X, f(X), a, lamda) < 1e-4:
                    break

            else:
                LOG.warning("No suitable exponential decomposition has been found"
                            "for dimless_omega=%.2e and dimless_wavenumber=%.2e",
                            dimensionless_omega, dimensionless_wavenumber)

        elif self.finite_depth_prony_decomposition_method.lower() == 'fortran':
            lamda, a, nexp = Delhommeau_f90.old_prony_decomposition.lisc(dimensionless_omega, dimensionless_wavenumber)
            lamda = lamda[:nexp]
            a = a[:nexp]

        else:
            raise ValueError("Unrecognized method name for the Prony decomposition.")

        # Add one more exponential function (actually a constant).
        # It is not clear where it comes from exactly in the theory...
        a = np.concatenate([a, np.array([2])])
        lamda = np.concatenate([lamda, np.array([0.0])])

        return a, lamda

    def evaluate(self, mesh1, mesh2, free_surface=0.0, sea_bottom=-np.infty, wavenumber=1.0):
        r"""The main method of the class, called by the engine to assemble the influence matrices.

        Parameters
        ----------
        mesh1: Mesh or CollectionOfMeshes
            mesh of the receiving body (where the potential is measured)
        mesh2: Mesh or CollectionOfMeshes
            mesh of the source body (over which the source distribution is integrated)
        free_surface: float, optional
            position of the free surface (default: :math:`z = 0`)
        sea_bottom: float, optional
            position of the sea bottom (default: :math:`z = -\infty`)
        wavenumber: float, optional
            wavenumber (default: 1.0)

        Returns
        -------
        tuple of numpy arrays
            the matrices :math:`S` and :math:`K`
        """

        depth = free_surface - sea_bottom
        if free_surface == np.infty: # No free surface, only a single Rankine source term

            a_exp, lamda_exp = np.empty(1), np.empty(1)  # Dummy arrays that won't actually be used by the fortran code.

            coeffs = np.array((1.0, 0.0, 0.0))

        elif depth == np.infty:

            a_exp, lamda_exp = np.empty(1), np.empty(1)  # Idem

            if wavenumber == 0.0:
                coeffs = np.array((1.0, 1.0, 0.0))
            elif wavenumber == np.infty:
                coeffs = np.array((1.0, -1.0, 0.0))
            else:
                coeffs = np.array((1.0, -1.0, 1.0))

        else:  # Finite depth
            a_exp, lamda_exp = self.find_best_exponential_decomposition(
                wavenumber*depth*np.tanh(wavenumber*depth),
                wavenumber*depth,
            )
            if wavenumber == 0.0:
                raise NotImplementedError
            elif wavenumber == np.infty:
                raise NotImplementedError
            else:
                coeffs = np.array((1.0, 1.0, 1.0))

        # Main call to Fortran code
        S, K = self.fortran_core.matrices.build_matrices(
            mesh1.faces_centers, mesh1.faces_normals,
            mesh2.vertices,      mesh2.faces + 1,
            mesh2.faces_centers, mesh2.faces_normals,
            mesh2.faces_areas,   mesh2.faces_radiuses,
            *mesh2.quadrature_points,
            wavenumber, depth,
            coeffs,
            self.tabulated_r_range, self.tabulated_z_range, self.tabulated_integrals,
            lamda_exp, a_exp,
            mesh1 is mesh2
        )

        if np.any(np.isnan(S)) or np.any(np.isnan(K)):
            raise RuntimeError("Green function returned a NaN in the interaction matrix.\n"
                    "It could be due to overlapping panels.")

        return S, K

################################

class XieDelhommeau(Delhommeau):
    """Variant of Nemoh's Green function, more accurate near the free surface.

    Same arguments and methods as :class:`Delhommeau`.
    """

    fortran_core = XieDelhommeau_f90



import logging

import numpy as np
import pandas as pd
import xarray as xr
from scipy.optimize import newton

from capytaine import __version__


LOG = logging.getLogger(__name__)

#######################
#  Import from Bemio  #
#######################

def dataframe_from_bemio(bemio_obj, wavenumber, wavelength):
    """Transform a :class:`bemio.data_structures.bem.HydrodynamicData` into a
        :class:`pandas.DataFrame`.

        Parameters
        ----------
        bemio_obj: Bemio data_stuctures.bem.HydrodynamicData class
            Loaded NEMOH, AQWA, or WAMIT data created using `bemio.io.nemoh.read`,
            `bemio.io.aqwa.read`, or `bemio.io.wamit.read` functions, respectively.
        wavenumber: bool
            If True, the coordinate 'wavenumber' will be added to the output dataset.
        wavelength: bool
            If True, the coordinate 'wavelength' will be added to the output dataset.
        """


    dofs = np.array(['Surge', 'Sway', 'Heave', 'Roll', 'Pitch', 'Yaw'])
    df = pd.DataFrame()
    for i in range(bemio_obj.body[0].num_bodies):
        difr_dict = []
        rad_dict = []

        rho = bemio_obj.body[0].rho
        g = bemio_obj.body[0].g

        if bemio_obj.body[i].water_depth == 'infinite':
            bemio_obj.body[i].water_depth = np.infty

        if bemio_obj.body[i].bem_code == 'WAMIT': # WAMIT coefficients need to be dimensionalized
            from_wamit = True

        for omega_idx, omega in enumerate(np.sort(bemio_obj.body[i].w)):

            # DiffractionProblem variable equivalents
            for dir_idx, dir in enumerate(bemio_obj.body[i].wave_dir):
                temp_dict = {}
                temp_dict['body_name'] = bemio_obj.body[i].name
                temp_dict['water_depth'] = bemio_obj.body[i].water_depth
                temp_dict['omega'] = omega
                temp_dict['rho'] = rho
                temp_dict['g'] = g
                temp_dict['wave_direction'] = np.radians(dir)
                temp_dict['convention'] = bemio_obj.body[i].bem_code
                temp_dict['influenced_dof'] = dofs
                
                if wavenumber or wavelength:
                    if temp_dict['water_depth'] == np.infty or omega**2*temp_dict['water_depth']/temp_dict['g'] > 20:
                        k = omega**2/temp_dict['g']
                    else:
                        k = newton(lambda x: x*np.tanh(x) - omega**2*temp_dict['water_depth']/temp_dict['g'], x0=1.0)/temp_dict['water_depth']
                    
                    if wavenumber:
                        temp_dict['wavenumber'] = k

                    if wavelength:
                        if k == 0.0:
                            temp_dict['wavelength'] = np.infty
                        else:
                            temp_dict['wavelength'] = 2*np.pi/k

                Fexc = np.empty(shape=bemio_obj.body[i].ex.re[:, dir_idx, omega_idx].shape, dtype=np.complex128)
                if from_wamit:
                    Fexc.real = bemio_obj.body[i].ex.re[:, dir_idx, omega_idx] * rho * g
                    Fexc.imag = bemio_obj.body[i].ex.im[:, dir_idx, omega_idx] * rho * g
                else:
                    Fexc.real = bemio_obj.body[i].ex.re[:, dir_idx, omega_idx]
                    Fexc.imag = bemio_obj.body[i].ex.im[:, dir_idx, omega_idx]
                temp_dict['diffraction_force'] = Fexc.flatten()
            
                try:
                    Fexc_fk = np.empty(shape=bemio_obj.body[i].ex.fk.re[:, dir_idx, omega_idx].shape, dtype=np.complex128)
                    if from_wamit:
                        Fexc_fk.real = bemio_obj.body[i].ex.fk.re[:, dir_idx, omega_idx] * rho * g
                        Fexc_fk.imag = bemio_obj.body[i].ex.fk.im[:, dir_idx, omega_idx] * rho * g
                    else:
                        Fexc_fk.real = bemio_obj.body[i].ex.fk.re[:, dir_idx, omega_idx]
                        Fexc_fk.imag = bemio_obj.body[i].ex.fk.im[:, dir_idx, omega_idx]
                    temp_dict['Froude_Krylov_force'] = Fexc_fk.flatten()

                except AttributeError:
                        # LOG.warning('\tNo Froude-Krylov forces found for ' + bemio_obj.body[i].name + ' at ' + str(dir) + \
                        #       ' degrees (omega = ' + str(omega) + '), replacing with zeros.')
                        temp_dict['Froude_Krylov_force'] = np.zeros((bemio_obj.body[i].ex.re[:, dir_idx, omega_idx].size,), dtype=np.complex128)

                difr_dict.append(temp_dict)

            # RadiationProblem + Hydrostatics variable equivalents
            for radiating_dof_idx, radiating_dof in enumerate(dofs):
                temp_dict = {}
                temp_dict['body_name'] = bemio_obj.body[i].name
                temp_dict['water_depth'] = bemio_obj.body[i].water_depth
                temp_dict['omega'] = omega
                temp_dict['rho'] = rho
                temp_dict['g'] = g
                temp_dict['influenced_dof'] = dofs
                temp_dict['radiating_dof'] = radiating_dof
                temp_dict['added_mass'] = bemio_obj.body[i].am.all[radiating_dof_idx, :, omega_idx].flatten()
                temp_dict['radiation_damping'] = bemio_obj.body[i].rd.all[radiating_dof_idx, :, omega_idx].flatten()

                if from_wamit:
                    temp_dict['added_mass'] = temp_dict['added_mass'] * rho
                    temp_dict['radiation_damping'] = temp_dict['radiation_damping'] * rho * omega

                if wavenumber or wavelength:
                    if temp_dict['water_depth'] == np.infty or omega**2*temp_dict['water_depth']/temp_dict['g'] > 20:
                        k = omega**2/temp_dict['g']
                    else:
                        k = newton(lambda x: x*np.tanh(x) - omega**2*temp_dict['water_depth']/temp_dict['g'], x0=1.0)/temp_dict['water_depth']
                    
                    if wavenumber:
                        temp_dict['wavenumber'] = k

                    if wavelength:
                        if k == 0.0:
                            temp_dict['wavelength'] = np.infty
                        else:
                            temp_dict['wavelength'] = 2*np.pi/k

                rad_dict.append(temp_dict)

    df = df.append(pd.DataFrame.from_dict(difr_dict).explode(['influenced_dof', 'diffraction_force', 'Froude_Krylov_force']))
    df = df.append(pd.DataFrame.from_dict(rad_dict).explode(['influenced_dof', 'added_mass', 'radiation_damping']))
    df = df.astype({'added_mass': np.float64, 'radiation_damping': np.float64, 'diffraction_force': np.complex128, 'Froude_Krylov_force': np.complex128})

    return df


#!/usr/bin/env python
#  -*- coding: utf-8 -*-
"""Functions to write mesh to different file formats.
Based on meshmagick <https://github.com/LHEEA/meshmagick> by François Rongère.
"""
# Copyright (C) 2017-2019 Matthieu Ancellin, based on the work of François Rongère
# See LICENSE file at <https://github.com/mancellin/capytaine>

import os
import time
import numpy as np

from capytaine.tools.optional_imports import import_optional_dependency


def write_mesh(filename, vertices, faces, file_format):
    """Driver function that writes every mesh file file_format known by meshmagick

    Parameters
    ----------
    filename: str
        name of the mesh file to be written on disk
    vertices: ndarray
        numpy array of the coordinates of the mesh's nodes
    faces: ndarray
        numpy array of the faces' nodes connectivities
    file_format: str
        file_format of the mesh defined in the extension_dict dictionary
    """

    if file_format not in extension_dict:
        raise IOError('Extension "%s" is not known' % file_format)

    writer = extension_dict[file_format]

    writer(filename, vertices, faces)


def write_DAT(filename, vertices, faces):
    """Writes .DAT file format for the DIODORE (PRINCIPA (c)) software.

    It also displays suggestions for inclusion into the .INP configuration
    file.

    Parameters
    ----------
    filename: str
        name of the mesh file to be written on disk
    vertices: ndarray
        numpy array of the coordinates of the mesh's nodes
    faces: ndarray
        numpy array of the faces' nodes connectivities
    """

    root_filename, ext = os.path.splitext(filename)
    filename = root_filename + ext.upper()
    ofile = open(filename, 'w')

    ofile.write('$\n$ Data for DIODORE input file : {0}\n'.format(root_filename.upper()))
    ofile.write('$ GENERATED BY MESHMAGICK ON {0}\n$\n'.format(time.strftime('%c')))

    ofile.write('$ NODE\n')
    vertex_block = \
        ''.join(
            (
                '\n'.join(
                    ''.join(
                        (
                            '{:8d}'.format(idx+1),
                            ''.join('{:13.5E}'.format(elt) for elt in node)
                        )
                    ) for (idx, node) in enumerate(vertices)
                ),

                '\n*RETURN\n'
            )
        )
    ofile.write(vertex_block)

    quad_block = '$\n$ ELEMENT,TYPE=Q4C000,ELSTRUCTURE={0}'.format(root_filename.upper())
    tri_block = '$\n$ ELEMENT,TYPE=T3C000,ELSTRUCTURE={0}'.format(root_filename.upper())
    nq = 0
    nt = 0
    for (idx, cell) in enumerate(faces+1):
        if cell[0] != cell[-1]:
            # quadrangle
            nq += 1
            quad_block = ''.join(
                (
                    quad_block,
                    '\n',
                    '{:8d}'.format(idx+1),
                    ''.join('{:8d}'.format(node_id) for node_id in cell)
                )
            )

        else:
            # Triangle
            nt += 1
            tri_block = ''.join(
                (
                    tri_block,
                    '\n',
                    '{:8d}'.format(idx+1),
                    ''.join('{:8d}'.format(node_id) for node_id in cell[:3])
                )
            )

    print('-------------------------------------------------')
    print('Suggestion for .inp DIODORE input file :')
    print('')
    print('*NODE,INPUT={0},FRAME=???'.format(root_filename))

    if nq > 0:
        quad_block = ''.join((quad_block, '\n*RETURN\n'))
        ofile.write(quad_block)
        print('*ELEMENT,TYPE=Q4C000,ELSTRUCTURE={0},INPUT={0}'.format(root_filename))
    if nt > 0:
        tri_block = ''.join((tri_block, '\n*RETURN\n'))
        ofile.write(tri_block)
        print('*ELEMENT,TYPE=T3C000,ELSTRUCTURE={0},INPUT={0}'.format(root_filename))

    print('')
    print('-------------------------------------------------')
    ofile.close()


def write_HST(filename, vertices, faces):
    """Writes .HST file format for the HYDROSTAR (Bureau Veritas (c)) software.

    Parameters
    ----------
    filename: str
        name of the mesh file to be written on disk
    vertices: ndarray
        numpy array of the coordinates of the mesh's nodes
    faces: ndarray
        numpy array of the faces' nodes connectivities
    """
    # TODO: allow many bodies

    ofile = open(filename, 'w')

    ofile.write(''.join((
        'PROJECT:\n',
        'USERS:   meshmagick\n\n'
        'NBODY   1\n'
        'RHO   1025.0\n'
        'GRAVITY   9.81\n\n'
    )))

    coordinates_block = ''.join((  # block
            'COORDINATES\n',
            '\n'.join(  # line
                ''.join(
                    (
                        '{:10d}'.format(idx+1),  # index
                        ''.join('{:16.6E}'.format(elt) for elt in node)  # node coordinates
                    )
                ) for (idx, node) in enumerate(vertices)
            ),
            '\nENDCOORDINATES\n\n'
    ))

    ofile.write(coordinates_block)

    cells_coordinates = ''.join((  # block
        'PANEL TYPE 0\n',
        '\n'.join(  # line
            ''.join(
                '{:10d}'.format(node_idx) for node_idx in cell
            ) for cell in faces + 1
        ),
        '\nENDPANEL\n\n'
    ))

    ofile.write(cells_coordinates)

    ofile.write('ENDFILE\n')

    ofile.close()


def write_TEC(filename, vertices, faces):
    """Writes .TEC file format for the TECPLOT (Tecplot (c)) visualisation software.

    Parameters
    ----------
    filename: str
        name of the mesh file to be written on disk
    vertices: ndarray
        numpy array of the coordinates of the mesh's nodes
    faces: ndarray
        numpy array of the faces' nodes connectivities
    """

    ofile = open(filename, 'w')

    nv = vertices.shape[0]
    nf = faces.shape[0]

    ofile.write('TITLE = \" THIS FILE WAS GENERATED BY MESHMAGICK - FICHIER : {} \" \n'.format(filename))

    ofile.write('VARIABLES = \"X\",\"Y\",\"Z\" \n')
    ofile.write('ZONE T=\"MESH\" \n')
    ofile.write('N={nv:10d} ,E={nf:10d} , F=FEPOINT, ET=QUADRILATERAL\n'.format(nv=nv, nf=nf))

    node_block = '\n'.join( # block
        ''.join(
            ''.join('{:16.6E}'.format(elt) for elt in node)
        ) for node in vertices
    ) + '\n'
    ofile.write(node_block)

    cells_block = '\n'.join(  # block
        ''.join(
            ''.join('{:10d}'.format(node_id) for node_id in cell)
        ) for cell in faces + 1
    ) + '\n'
    ofile.write(cells_block)

    ofile.close()

    return 1


def write_VTU(filename, vertices, faces):
    """Writes .vtu file format for the paraview (Kitware (c)) visualisation software.

    It relies on the VTK library for its writer. VTU files use the last XML file format of the VTK library.

    Parameters
    ----------
    filename: str
        name of the mesh file to be written on disk
    vertices: ndarray
        numpy array of the coordinates of the mesh's nodes
    faces: ndarray
        numpy array of the faces' nodes connectivities
    """
    vtk = import_optional_dependency("vtk")

    writer = vtk.vtkXMLUnstructuredGridWriter()
    writer.SetDataModeToAscii()
    writer.SetFileName(str(filename))

    unstructured_grid = _build_vtkUnstructuredGrid(vertices, faces)
    writer.SetInputData(unstructured_grid)
    writer.Write()


def write_VTP(filename, vertices, faces):
    """Writes .vtp file format for the Paraview (Kitware (c)) visualisation software.

    It relies on the VTK library for its writer. VTP files use the last XML file format of the VTK library and
    correspond to polydata.

    Parameters
    ----------
    filename: str
        name of the mesh file to be written on disk
    vertices: ndarray
        numpy array of the coordinates of the mesh's nodes
    faces: ndarray
        numpy array of the faces' nodes connectivities
    """

    vtk = import_optional_dependency("vtk")

    writer = vtk.vtkXMLPolyDataWriter()
    writer.SetDataModeToAscii()
    writer.SetFileName(str(filename))

    polydata = _build_vtkPolyData(vertices, faces)
    writer.SetInputData(polydata)
    writer.Write()


def write_VTK(filename, vertices, faces):
    """Writes .vtk file format for the Paraview (Kitware (c)) visualisation software.

    It relies on the VTK library for its writer. VTK files use the legagy ASCII file format of the VTK library.

    Parameters
    ----------
    filename: str
        name of the mesh file to be written on disk
    vertices: ndarray
        numpy array of the coordinates of the mesh's nodes
    faces: ndarray
        numpy array of the faces' nodes connectivities
    """

    nv = vertices.shape[0]
    nf = faces.shape[0]

    triangle_mask = (faces[:, 0] == faces[:, -1])
    quadrangles_mask = np.invert(triangle_mask)
    nb_triangles = len(np.where(triangle_mask)[0])
    nb_quandrangles = len(np.where(quadrangles_mask)[0])

    with open(filename, 'w') as f:

        f.write('# vtk DataFile Version 4.0\n')
        f.write('vtk file generated by meshmagick on %s\n' % time.strftime('%c'))
        f.write('ASCII\n')
        f.write('DATASET POLYDATA\n')
        f.write('POINTS %u float\n' % nv)

        for vertex in vertices:
            f.write('%f %f %f\n' % (vertex[0], vertex[1], vertex[2]))

        f.write('POLYGONS %u %u\n' % (nf, 4*nb_triangles+5*nb_quandrangles))

        for face in faces:
            if face[0] == face[-1]:  # Triangle
                f.write('3 %u %u %u\n' % (face[0], face[1], face[2]))
            else:  # Quadrangle
                f.write('4 %u %u %u %u\n' % (face[0], face[1], face[2], face[3]))


def _build_vtkUnstructuredGrid(vertices, faces):
    """Internal function that builds a VTK object for manipulation by the VTK library.

    Parameters
    ----------
    vertices: ndarray
        numpy array of the coordinates of the mesh's nodes
    faces: ndarray
        numpy array of the faces' nodes connectivities

    Returns
    -------
    vtkObject
    """

    vtk = import_optional_dependency("vtk")

    nv = max(np.shape(vertices))
    nf = max(np.shape(faces))

    vtk_mesh = vtk.vtkUnstructuredGrid()
    vtk_mesh.Allocate(nf, nf)

    # Building the vtkPoints data structure
    vtk_points = vtk.vtkPoints()
    vtk_points.SetNumberOfPoints(nv)
    for idx, vertex in enumerate(vertices):
        vtk_points.SetPoint(idx, vertex)

    vtk_mesh.SetPoints(vtk_points)  # Storing the points into vtk_mesh

    # Building the vtkCell data structure
    for cell in faces:
        if cell[-1] in cell[:-1]:
            vtk_cell = vtk.vtkTriangle()
            nc = 3
        else:
            # #print 'quadrangle'
            vtk_cell = vtk.vtkQuad()
            nc = 4

        for k in range(nc):
            vtk_cell.GetPointIds().SetId(k, cell[k])

        vtk_mesh.InsertNextCell(vtk_cell.GetCellType(), vtk_cell.GetPointIds())
    return vtk_mesh


def _build_vtkPolyData(vertices, faces):
    """Builds a vtkPolyData object from vertices and faces"""

    vtk = import_optional_dependency("vtk")

    # Create a vtkPoints object and store the points in it
    points = vtk.vtkPoints()
    for point in vertices:
        points.InsertNextPoint(point)

    # Create a vtkCellArray to store faces
    cell_array = vtk.vtkCellArray()
    for face_ids in faces:
        if face_ids[0] == face_ids[-1]:
            # Triangle
            curface = face_ids[:3]
            vtk_face = vtk.vtkTriangle()
        else:
            # Quadrangle
            curface = face_ids[:4]
            vtk_face = vtk.vtkQuad()

        for idx, id in enumerate(curface):
            vtk_face.GetPointIds().SetId(idx, id)

        cell_array.InsertNextCell(vtk_face)

    polydata_mesh = vtk.vtkPolyData()
    polydata_mesh.SetPoints(points)
    polydata_mesh.SetPolys(cell_array)

    return polydata_mesh


def write_NAT(filename, vertices, faces):
    """Writes .nat file format as defined into the load_NAT function.

    Parameters
    ----------
    filename: str
        name of the mesh file to be written on disk
    vertices: ndarray
        numpy array of the coordinates of the mesh's nodes
    faces: ndarray
        numpy array of the faces' nodes connectivities

    See Also
    --------
    load_NAT
    """

    ofile = open(filename, 'w')

    nv = max(np.shape(vertices))
    nf = max(np.shape(faces))

    ofile.write('%6u%6u\n' % (0, 0))  # lire les symmetries dans args...
    ofile.write('%6u%6u\n' % (nv, nf))
    for vertex in vertices:
        ofile.write('%15.6E%15.6E%15.6E\n' % (vertex[0], vertex[1], vertex[2]))
    for cell in faces+1:
        ofile.write('%10u%10u%10u%10u\n' % (cell[0], cell[1], cell[2], cell[3]))

    ofile.close()


def write_NEM(filename, vertices, faces):
    """Writes mesh files used by the Mesh tool included in Nemoh

    Parameters
    ----------
    filename : str
        name of the mesh file to be written on disk
    vertices: ndarray
        numpy array of the coordinates of the mesh's nodes
    faces: ndarray
        numpy array of the faces' nodes connectivities

    Note
    ----
    This file format is different from that used by Nemoh itself. It is only used by the Mesh tool.
    """
    ofile = open(filename, 'w')

    ofile.write('%u\n' % vertices.shape[0])
    ofile.write('%u\n' % faces.shape[0])

    for vertex in vertices:
        ofile.write('%15.6f\t%15.6f\t%15.6f\n' % (vertex[0], vertex[1], vertex[2]))

    for face in faces+1:
        ofile.write('%10u\t%10u\t%10u\t%10u\n' % (face[0], face[1], face[2], face[3]))

    ofile.close()


def write_GDF(filename, vertices, faces):
    """Writes .gdf file format for the WAMIT (Wamit INC. (c)) BEM software.

    Parameters
    ----------
    filename: str
        name of the mesh file to be written on disk
    vertices: ndarray
        numpy array of the coordinates of the mesh's nodes
    faces: ndarray
        numpy array of the faces' nodes connectivities
    """

    nf = max(np.shape(faces))

    ofile = open(filename, 'w')

    ofile.write('GDF file generated by meshmagick on %s\n' % time.strftime('%c'))

    ofile.write('%16.6f%16.6f\n' % (100.0, 9.81))
    ofile.write('%12u%12u\n' % (0, 1))  # TODO : mettre les symetries en argument
    ofile.write('%12u\n' % nf)

    for cell in faces:
        for k in range(4):
            cur_vertices = vertices[cell[k], :]
            ofile.write('%16.6E%16.6E%16.6E\n' % (cur_vertices[0], cur_vertices[1], cur_vertices[2]))

    ofile.close()


def write_MAR(filename, vertices, faces):
    """Writes mesh files to be used with Nemoh BEM software (Ecole Centrale de Nantes)

    Parameters
    ----------
    filename: str
        name of the mesh file to be written on disk
    vertices: ndarray
        numpy array of the coordinates of the mesh's nodes
    faces: ndarray
        numpy array of the faces' nodes connectivities
    """

    # TODO: detect symmetry in Oxz plane

    ofile = open(filename, 'w')

    ofile.write('{0:6d}{1:6d}\n'.format(2, 0))  # TODO : mettre les symetries en argument

    for (idx, vertex) in enumerate(vertices):
        ofile.write('{0:6d}{1:16.6f}{2:16.6f}{3:16.6f}\n'.format(idx+1, vertex[0], vertex[1], vertex[2]))

    ofile.write('{0:6d}{1:6d}{2:6d}{3:6d}{4:6d}\n'.format(0, 0, 0, 0, 0))

    cell_block = '\n'.join(
        ''.join('{0:10d}'.format(elt) for elt in cell)
        for cell in faces + 1
    ) + '\n'
    ofile.write(cell_block)
    ofile.write('%6u%6u%6u%6u\n' % (0, 0, 0, 0))

    ofile.close()

    print('WARNING: if you described only one part of the mesh using symmetry for Nemoh, you may manually modify the ' \
          'file header accordingly')


def write_RAD(filename, vertices, faces):
    raise NotImplementedError


def write_STL(filename, vertices, faces):
    """Writes .stl file format. It relies on the VTK library for its writer.

    Parameters
    ----------
    filename: str
        name of the mesh file to be written on disk
    vertices: ndarray
        numpy array of the coordinates of the mesh's nodes
    faces: ndarray
        numpy array of the faces' nodes connectivities
    """

    # TODO : replace this implementation by using the vtk functionalities

    # Triangulating quads
    t1 = (0, 1, 2)
    t2 = (0, 2, 3)

    quads_ids = np.where(faces[:, 0] != faces[:, -1])[0]

    new_faces = faces[quads_ids].copy()
    new_faces[:, :3] = new_faces[:, t1]
    new_faces[:, -1] = new_faces[:, 0]

    faces[quads_ids, :3] = faces[:, t2][quads_ids]
    faces[quads_ids, -1] = faces[quads_ids, 0]

    faces = np.concatenate((faces, new_faces))

    # Writing file
    ofile = open(filename, 'w')

    ofile.write('solid meshmagick\n')

    for face in faces:
        if face[0] != face[3]:
            raise RuntimeError("""Only full triangle meshes are accepted in STL files.
              Please consider using the --triangulate-quadrangles option (-tq) to
              perform a prior triangulation of the mesh""")

        # Computing normal
        v0 = vertices[face[0], :]
        v1 = vertices[face[1], :]
        v2 = vertices[face[2], :]

        n = np.cross(v1 - v0, v2 - v0)
        n /= np.linalg.norm(n)

        block_facet = ''.join(['  facet normal ', ''.join('%15.6e' % ni for ni in n) + '\n',
                               '    outer loop\n',
                               '      vertex', ''.join('%15.6e' % Vi for Vi in v0) + '\n',
                               '      vertex', ''.join('%15.6e' % Vi for Vi in v1) + '\n',
                               '      vertex', ''.join('%15.6e' % Vi for Vi in v2) + '\n',
                               '    endloop\n',
                               '  endfacet\n'])
        ofile.write(block_facet)
    ofile.write('endsolid meshmagick\n')
    ofile.close()


def write_INP(filename, vertices, faces):
    raise NotImplementedError('INP writer is not implementer yet')


def write_MSH(filename, vertices, faces):
    raise NotImplementedError('MSH writer is not implemented yet')


def write_MED(filename, vertices, faces):
    raise NotImplementedError('MED writer is not implemented yet')


def write_WRL(filename, vertices, faces):
    raise NotImplementedError('VRML writer is not implemented yet')


extension_dict = {  # keyword,  writer
    'mar': write_MAR,
    'nemoh': write_MAR,
    'wamit': write_GDF,
    'gdf': write_GDF,
    'diodore-inp': write_INP,
    'inp': write_INP,
    'diodore-dat': write_DAT,
    'hydrostar': write_HST,
    'hst': write_HST,
    'natural': write_NAT,
    'nat': write_NAT,
    'gmsh': write_MSH,
    'msh': write_MSH,
    'rad': write_RAD,
    'radioss': write_RAD,
    'stl': write_STL,
    'vtu': write_VTU,
    'vtp': write_VTP,
    'paraview-legacy': write_VTK,
    'vtk': write_VTK,
    'tecplot': write_TEC,
    'tec': write_TEC,
    'med': write_MED,
    'salome': write_MED,
    'vrml': write_WRL,
    'wrl': write_WRL,
    'nem': write_NEM,
    'nemoh_mesh': write_NEM
}
